import * as errors from "node_modules/errors.js/errors";

import * as utils from "./utilities";
import * as api from "./api";
import * as response from "./response";

export type HttpPromise = Promise<{} | any[] | string | null>;
export type HttpResponse = response.Response<any>;

export const ErrEmptyURL = new errors.ErrorString("cannot use an empty url");
export const ErrInvalidReq = new errors.ErrorString("invalid request type provided");

export const EmptyURLResp = new response.Response(null, ErrEmptyURL);
export const InvalidReqResp = new response.Response(null, ErrInvalidReq);

export const DefaultResponseType = "text";

const nullFn = (data: response.Response<any>) => { };

export function New(method: string, url: string, data: any, requestType: utils.Types, responseType: utils.Types): HttpPromise {
	return HttpRequest(method, url, data, requestType, responseType);
}

export function Get(url: string): HttpPromise {
	return HttpRequest("GET", url, null, utils.Types.JSON, utils.Types.JSON);
}

export function Post(url: string, data: any): HttpPromise {
	return HttpRequest("POST", url, data, utils.Types.JSON, utils.Types.JSON);
}

export function Put(url: string, data: any): HttpPromise {
	return HttpRequest("PUT", url, data, utils.Types.JSON, utils.Types.JSON);
}

export function Delete(url: string): HttpPromise {
	return HttpRequest("DELETE", url, null, utils.Types.JSON, utils.Types.JSON);
}

export function HttpRequest(method: string, url: string, data: any, requestType: utils.Types, responseType: utils.Types): HttpPromise {
	var fn: (data: HttpResponse) => void = nullFn;
	var p = <HttpPromise>new Promise((resolve, reject) => {
		fn = resolve;
	});

	if (!url) {
		fn(EmptyURLResp);
		return p;
	}

	if (!requestType) {
		fn(InvalidReqResp)
		return p;
	}

	let req = new XMLHttpRequest();
	// Send cookies to CORs domains
	req.withCredentials = true;
	// Ensure method is all upper-case
	method = method.toUpperCase();
	// Prepare and send the request
	req.open(method, utils.GetURL(url), true);
	// Set response type
	req.responseType = getResponseType(responseType);
	// Set onload listener
	req.onload = handleLoad;
	// Send request
	req.send(utils.GetRequestData(req, requestType, data));
	// Return a promise
	return p;

	function handleLoad() {
		if (req.readyState !== 4) {
			return;
		}

		fn(getData(req, responseType));
	}
}


function getData(req: XMLHttpRequest, rt: utils.Types): HttpResponse {
	if (rt == utils.Types.Text) {
		return new response.Response<any>(req.responseText, getStatusCodeError(req));
	}

	// Process response
	const nar = api.New(req.response);
	if (nar.Error !== null) {
		// Return response generated directly from response value
		return new response.Response<any>(req.response, getStatusCodeError(req));
	}

	// Assign api request const
	const ar = nar.Value;
	// Return response generated from api response
	return new response.Response<any>(ar.data, ar.errors);
}

function getStatusCodeError(req: XMLHttpRequest): errors.Error | null {
	if(req.status >= 200 && req.status < 400){
		return null;
	}

	return new errors.ErrorString(req.statusText);
}



function getResponseType(rt: utils.Types): XMLHttpRequestResponseType {
	return <XMLHttpRequestResponseType>getResponseTypeStr(rt);
}

function getResponseTypeStr(rt: utils.Types): string {
	switch (rt) {
		case utils.Types.JSON:
			return  "json";
		case utils.Types.ArrayBuffer:
			return "arraybuffer";
		case utils.Types.Blob:
			return "blob";
		case utils.Types.Document:
			return "document";
		case utils.Types.Text:
			return "text";

		default:
			console.error("Non-supported response type, reverting to text")
			return "text";
	}
}