import * as errors from "node_modules/errors.js/errors"

import * as utils from "./utilities";
import * as response from "./response";
import * as api from "./api";

export type JSONPPromise = Promise<JSONPResponse>;
export type JSONPResponse = response.Response<any>;

export const ErrTimedOut = new errors.ErrorString("request timed out");
export const ErrLoading = new errors.ErrorString("error while loading");

const prefix = "_jsonP_";
const nullFn = function () { };

export async function Get(url: string): JSONPPromise {
	return Request(url, null);
}

export async function Request(url: string, data: any): JSONPPromise {
	let fn: RequestFn;
	const p = new Promise<JSONPResponse>((resolve, reject) => {
		fn = resolve;
	});

	// Set key
	const key = getKey();
	// Set timeout key
	const tok = setTimeout(() => error(ErrTimedOut), 3000);
	// Create script
	const script = getScript(url, key, data);
	// Set error listener, because we do not have access to the response body nor the status code
	// during an unexpected error (EG 400, 404, 500), we have to return a non-descriptive ErrLoading
	script.onerror = () => error(ErrLoading);
	// Append script to document body
	document.body.appendChild(script);
	// Set window function
	(<any>window)[key] = success;
	// Return promise
	return p;

	function success(data: any) {
		exit();
		fn(getResponse(data));
	}

	function error(err: errors.Error) {
		exit();
		fn(new response.Response(null, err));
	}

	function exit() {
		clearTimeout(tok);
		removeScript(script);
		script.onerror = nullFn;
	}
}

export type RequestFn = (data: any) => void;

function getScript(src: string, key: string, payload: any): HTMLScriptElement {
	const script = <HTMLScriptElement>document.createElement("script");
	script.src = src;
	script.type = "application/javascript";

	const u = new URL(script.src);
	u.searchParams.set("callback", key)
	if (payload !== null) {
		u.searchParams.set("data", encodeURI(JSON.stringify(payload)));
	}

	script.src = u.href;
	return script;
}

function getKey(): string {
	return prefix + Date.now().toString() + getRandom(100000, 999999).toString()
}

function getRandom(min: number, max: number) {
	return Math.round(Math.random() * (max - min) + min);
}

function getResponse(data: any): JSONPResponse {
	const nar = api.New(data);
	if (nar.Error !== null) {
		return new response.Response<any>(data, null);
	}

	return new response.Response<any>(nar.Value.data, nar.Value.errors);
}

function removeScript(script: HTMLScriptElement) {
	let parent = script.parentElement;
	if (!parent) {
		return;
	}

	parent.removeChild(script);
}